from zhongwen.number import 發生例外則回覆非數常數
from pathlib import Path
import pandas as pd
import logging
logger = logging.getLogger(Path(__file__).stem)

@發生例外則回覆非數常數
def 連續消長次數(時序資料):
    s = 時序資料
    d = s - s.shift()
    return 最近連續正負次數(d)
    
@發生例外則回覆非數常數
def 最近連續非零正數次數(時序):
    import numpy as np
    a = np.sign(時序)
    s = 0
    for n in reversed(a):
        if np.isnan(n) or n <= 0: return s
        s+=n
    return s

@發生例外則回覆非數常數
def 最近連續正負次數(時序) -> int:
    import numpy as np
    a = np.sign(時序).dropna()
    s = 0
    for n in a[::-1]:
        if np.isnan(n): return int(s)
        if s != 0 and s*n < 0: return int(s)
        s+=n
    return int(s)

def 趨勢分析(時序, 時序數下限=3, 數據名稱=None, 匯出圖檔=None, 圖示=None, 時序周期="年", 數值單位='元'):
    '回傳趨勢及r方數之數組，如時序數低於下限則回傳(0, 0)'
    from fhopecc.洄瀾打狗人札記 import 網站本地目錄
    from sklearn.linear_model import LinearRegression
    from sklearn.metrics import r2_score
    import matplotlib.pyplot as plt
    import numpy as np
    import pandas as pd
    s = 時序
    if not isinstance(s, pd.Series):
        raise ValueError('時序參數型態需為 pd.Series，而傳入型態為{type(時序)}')
    if len(s) < 時序數下限:
        return (0, 0)
    if pd.api.types.is_integer_dtype(s.index):
        s = 序列缺項補零(s)
    X = np.arange(1, len(s)+1).reshape(-1, 1)
    Y = s
    reg = LinearRegression().fit(X, Y)
    std = Y.std()
    趨勢 = reg.coef_[0]/std
    r2 =  r2_score(Y, reg.predict(X))
    if 匯出圖檔 or 圖示:
        title = f'{數據名稱}趨勢圖'
        plt.rcParams['font.sans-serif'] = ['Microsoft YaHei'] 
        plt.title(f'{title}(成長率：{趨勢:.2%}、r方：{r2:.2%})')
        plt.xlabel(f'{時序周期}數')
        plt.ylabel(f'{數值單位}')
        plt.scatter(X, Y)
        line = [reg.coef_* x for x in X] + reg.intercept_
        plt.plot(X, line, color='red')
        if 匯出圖檔:
            png = f'{title}.png'
            plt.savefig(網站本地目錄 / 'images' / png)
            logger.info(f'匯出{png}')
        if 圖示:
            plt.show()
        plt.close()
    return (趨勢, r2)

def 數據周期(數據, 發布周期='年', 圖示=None, 數據名稱=None, 頻譜圖=False):
    '波動幅度未達直流分量三成者或數據時長未大於周期1.5倍者，視作無周期'
    import matplotlib.pyplot as plt
    import numpy as np
    import pandas as pd
    from fhopecc.洄瀾打狗人札記 import 網站本地目錄
    match 數據:
        case pd.Series():
            samples = 數據.tolist()
        case _:
            samples = list(數據)

    if len(samples) < 5: return 0

    y1 = samples
    x1 = np.arange(0, len(samples))
   
    fft = np.fft.fft(samples) #[:int(len(samples)/2)]
    y2 = np.abs(fft)/(int(len(fft)/2)) # 頻幅譜
    x2 = np.fft.fftfreq(len(y2), 1)
    y2[0] = y2[0]/2 # 頻率0表垂直位移，即直流分量，因正負0各疊加一次，須再除以2還原。
   
    argmax = np.argmax(y2[1:]) + 1 # 直流分量不計
    peak = y2[argmax] # 峰值
    if peak / y2[0] < 0.3: # 波動幅度未達直流分量三成者，視作無周期
        return 0
    f = x2[argmax] # 峰值頻率
    y3 = np.angle(fft) # 頻相譜，相位為應變項
    phase = y3[argmax] # 峰值相位

    if 數據名稱:
        plt.rcParams['font.sans-serif'] = ['Microsoft JhengHei'] 
        plt.rcParams['axes.unicode_minus'] = False
        if 頻譜圖:
            p1 = plt.subplot(211)
            p2 = plt.subplot(212)
        else:
            p1 = plt.subplot(111)

        p1.scatter(x1, y1)
        for i, (xi, yi) in enumerate(zip(x1, y1)):
            p1.text(xi, yi, f'{yi:.2f}', ha='center', va='bottom')
        p1.set_xticks(x1)
        p1.set_xlabel(f'{發布周期}數')

        x2h = x2[:int(len(x2)/2)+1]
        y2h = y2[:int(len(x2)/2)+1]

        if 頻譜圖:
            p2.scatter(x2h, y2h)
            for i, (xi, yi) in enumerate(zip(x2h, y2h)):
                p2.text(xi, yi, f'{yi:.2f}', ha='center', va='bottom')
            p2.set_xticks(x2h)
            p2.set_xticklabels([f'{1/f:.2f}' for f in x2h], rotation=90)
            p2.set_xlabel(f'周期({發布周期})')

        p1.axhline(y=y2h[0], color='r', linestyle='-')
        x4 = np.linspace(min(x1), max(x1), 100)
        y4 = y2[0] + peak*np.cos(f*2*np.pi*x4 + phase) # 峰值頻率餘弦諧波
        p1.plot(x4, y4, color='orange')
        title = f'{數據名稱}周期圖\n'
        title += f"周期為{1/f:.2f}{發布周期}，平均{y2h[0]:.2f}元，上下振盪{peak:.2f}元"
        p1.set_title(title)
        if 數據名稱:
            png = f'{數據名稱}周期圖.png'
            plt.savefig(網站本地目錄 / 'images' / png)
            plt.close()
        if 圖示:
            plt.show()

    # 數據時長未大於周期1.5倍者，len/(1/f) = len*f
    if len(samples) * f  < 1.5: return 0
    return 1/f

def 正弦離散轉換():
    import matplotlib.pyplot as plt
    import numpy as np
    plt.rcParams['font.sans-serif'] = ['Microsoft JhengHei'] 
    plt.rcParams['axes.unicode_minus'] = False

    x1 = np.linspace(0, 1, 30)   
    y1 = 0.7 + np.cos(3*2*np.pi*x1+7)
    # plt.scatter(x, y)

    fft = np.fft.fft(y1)
    y = np.abs(fft) / (30/2)
    p = np.angle(fft)
    print(p)
    y[0] = y[0]/2 # 頻率0表直流分量，即垂宜位移，正負0各疊加一次，須再除以2還原。
    # x = np.linspace(0, 30-1, 30) # f_s = 30Hz, 頻率間隔 = 樣頻/樣數 = 30/30 = 1 Hz
    x = np.fft.fftfreq(30, 1/30)
    argmax = y.argmax()
    print(x[argmax])
    d = y[0]
    a = y[argmax]
    f = x[argmax]
    p = p[argmax]
    print(f'{d} + np.cos({f}*2*pi*x1+{p})')
    # plt.scatter(x, y)
    # plt.xticks(x, rotation=90)
    y2 = d + np.cos(f*2*np.pi*x1+p)
    plt.scatter(x1, y1)
    plt.scatter(x1, y2)
    plt.show()

def 最近缺失序列(序列:pd.Series):
    '找出最大序列索引缺項，即不連續缺口，並回傳索引值及序位'
    s = 序列.sort_index()
    full_index = pd.Index(range(s.index[0], s.index[-1]))
    # 找出缺項
    missing_index = full_index.difference(s.index)
    if not missing_index.empty:
        return (missing_index[-1], s.index.max()-missing_index[-1])
    else:
        return (None, 0)

def 序列缺項補零(序列:pd.Series):
    '序列索引缺項，即不連續缺口補零'
    s = 序列.sort_index()
    full_index = pd.Index(range(s.index[0], s.index[-1]))
    # 找出缺項
    missing_index = full_index.difference(s.index)
    # 將缺項插入序列並賦值為 0
    for miss in missing_index:
        s.loc[miss] = 0
    # 按索引順序重排
    return s.sort_index()

def 繪季節解析圖(時序):
    from statsmodels.tsa.seasonal import seasonal_decompose
    ts = 時序 
    r = seasonal_decompose(ts, period=12)
    r.plot()
    return r

class 無法於一階差分內穩定(Exception):pass
def 取穩定階數(時序):
    '回傳是否穩定及穩定差分階數'
    from 股票分析.股票基本資料分析 import 查股票代號
    from statsmodels.tsa.stattools import adfuller
    ts = 時序
    _, pvalue, *_ = adfuller(ts)
    if pvalue < 0.05:
        return 0
    else:
        _, pvalue, *_ = adfuller(ts.diff().dropna())
        if pvalue < 0.05:
            return 1
        raise 無法於一階差分內穩定()

if __name__ == '__main__':
    pass
    from 股票分析.股票基本資料分析 import 查股票代號
    from 股票分析.股利分析 import 年度股利分派表
    from zhongwen.date import 今日, 上年初
    import pandas as pd
    股票代號 = 查股票代號('台積電') # 潤弘及志超標準景氣循環股
    s = 年度股利分派表().loc[股票代號]
    if 今日().month < 4: # 年報尚未公布，故上年度股利可能未發放
        s = s[s.index < pd.to_datetime(上年初())]
    else:
        s = s[s.index < pd.to_datetime(今日())] # 歸屬今年度股利可能尚未完全發放，故排除。
    # t = 數據周期(s, 圖示=True)
    # breakpoint()
    a, b = 趨勢分析(s.配息, 圖示=True, 數據名稱='配息')
    print(a, b)



from pathlib import Path
from diskcache import Cache
from bs4 import BeautifulSoup as bs
import re
cache = Cache(Path.home() / 'cache' / Path(__file__).name)

@cache.memoize(expire=30*24*60*60)
def 抓取法規(url):
    import requests
    url = 'https://law.nfa.gov.tw/mobile/law.aspx?LSID=FL005010'
    user_agent = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/106.0.0.0 Safari/537.36 Edg/106.0.1370.52'
    response = requests.get(url, headers={ 'user-agent': user_agent })
    law = bs(response.text, 'lxml')
    law = law.select_one("#contenter")
    law = law.text
    pat = r'([一二三四五六七八九十]+)\s+'
    law = re.sub(pat, r'\1、', law)
    pat = r'\)\s+'
    law = re.sub(pat, r')', law)
    pat = r'\s+|檢視歷史法規'
    law = re.sub(pat, r'', law)
    return law

def 法規類別():
    from zhongwen.file import 抓取
    html = 抓取('https://law.nfa.gov.tw/mobile/category.aspx?typecode=A001')
    print(html)



from zhongwen.number import 半型數字表, 全型數字表, 小寫數字表, 小寫位名表
from zhongwen.number import 大寫數字表, 大写数字表, 大寫位名表, 組名表, 组名表
from zhongwen.number import 最簡約數
全數字表  = 半型數字表 + 全型數字表 + 小寫數字表 + 小寫位名表
全數字表 += 大寫數字表 + 大写数字表 + 大寫位名表 + 組名表 + 组名表

def 取數值(n, 全取=False):
    from zhongwen.number import 轉數值
    import re
    pat = f'[{全數字表}]+[點.]?[{全數字表}]*[%]?'

    if ms:=re.findall(pat, n):
        ps = [轉數值(m) for m in ms]
   
    if 全取: return ps
    
    try:
        return ps[0]
    except UnboundLocalError:
        raise UnboundLocalError(f"'{n}'無法解析為數值字串！")

    return 轉數值(n)



import unittest

class Test(unittest.TestCase):
    def test(self):
        from zhongwen.數 import 取數值

        self.assertEqual(取數值('中間一二三數字'), 123)
        self.assertEqual(取數值('中間一二三數字246'), 123)
        self.assertEqual(取數值('中間一二三數字2.46', 全取=True), [123, 2.46])


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.INFO)
    logging.getLogger('googleclient').setLevel(logging.CRITICAL)
    logging.getLogger('matplotlib').setLevel(logging.CRITICAL)
    logging.getLogger('faker').setLevel(logging.CRITICAL)
    unittest.main()
    suite = unittest.TestSuite()
    suite.addTest(Test('test'))  # 指定測試
    unittest.TextTestRunner().run(suite)



def 同字首詞(字首, 字典):
    from pygtrie import PrefixSet
    from zhongwen.text import 是否為字元
    trie = PrefixSet(字典)
    suggests = []
    comps = []
    for i in range(0, len(字首)):
        prefix = 字首[-1*i-1:]
        if not 是否為字元(prefix[0]): 
            break
        s = [''.join(k) for k in trie.iter(prefix)]
        suggests = s+suggests
        comps = [c[len(prefix):] for c in s] + comps
    return [{'word':word, 'abbr':abbr, 'kind':'K'} for (word, abbr) in zip(comps, suggests)]





def 提出請求(請求事項:str):
    import os
    r = 請求事項
    if '法規' in r:
        r = r.replace('法規', '')
        cmd = f'start https://glrslaw.e-land.gov.tw/SearchAllResultList.aspx?KW={r}'
        os.system(cmd)

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("-q", "--query", help="提出請求")
    args = parser.parse_args()
    if args.query:
        提出請求(args.query)



from zhongwen.pandas_tools import 可顯示
from pathlib import Path
from diskcache import Cache
import pandas as pd
cache = Cache(Path.home() / 'cache' / Path(__file__).stem)

def 中央法律():
    url = 'https://sendlaw.moj.gov.tw/PublicData/GetFile.ashx?DType=XML&AuData=CF'
    z = Path.home() / 'TEMP' / 'falv.zip'
    下載(url, z)
    falv = Path.home() / 'TEMP' / 'law' / 'FalV.xml'
    解壓(z, falv.parent)
    from xml.etree.cElementTree import parse
    return parse(falv)

def 中央命令():
    url = 'https://sendlaw.moj.gov.tw/PublicData/GetFile.ashx?DType=XML&AuData=CM'
    z = Path.home() / 'TEMP' / 'mingling.zip'
    下載(url, z)
    minling = Path.home() / 'TEMP' / 'law' / 'MingLing.xml'
    解壓(z, minling.parent)
    from xml.etree.cElementTree import parse
    return parse(minling)

@可顯示
@cache.memoize(expire=30*24*60*60)
def 法規條文() -> pd.DataFrame:
    '法規條文係以法規名稱、異動日期及條號為索引來查詢條文內容'
    from pandas import DataFrame, concat
    l = 中央法律()
    def 法條資料表(法規):
        def 條號數(s):
            pat = r'第[^\d]*([-\d]+)[^\d]*條'
            if m:=re.match(pat, s):
                return m[1]
            return s
        條號 = [條號數(n.find('條號').text) for n in 法規.findall('*/條文')]
        條文內容 = [n.find('條文內容').text for n in 法規.findall('*/條文')]
        df = DataFrame(data = {'條號':條號, '條文內容':條文內容})
        df['法規名稱'] = 法規.find('法規名稱').text
        df['異動日期'] = 法規.find('最新異動日期')
        return df[['法規名稱', '異動日期', '條號', '條文內容']]
    law = concat([法條資料表(法規) for 法規 in 中央法律().findall('法規')])
    mingling = concat([法條資料表(法規) for 法規 in 中央命令().findall('法規')]) 
    df = concat([law, mingling])
    from zhongwen.date import 取日期
    df['異動日期'] = df.異動日期.map(取日期)
    return df

def 產製法規文字檔(法規名稱):
    df = 法規條文()
    df = df.query('法規名稱==@法規名稱')[["條號", "條文內容"]]
    content = '\n'.join([f"第{r.條號}條{r.條文內容}"for r in df.itertuples()])
    txt = Path(f'{法規名稱}.txt')
    txt.write_text(content)

if __name__ == '__main__':
    法規條文(顯示=True)



from invoke import task
from pathlib import Path

@task(default=True)
def setup_dev(c):
    from fhopecc import env
    import os
    github = env.githubdrive / 'github'
    p = Path(__file__).parent
    pythonpath = ";".join([str(Path(__file__).parent), os.environ['PYTHONPATH']])
    os.system(f'setx PYTHONPATH {pythonpath}')

@task
def hanlp(c):
    from pathlib import Path
    from zhongwen.file import 下載
    import logging
    logging.basicConfig(level=logging.INFO)
    local = Path.home() / 'AppData\Roaming\hanlp'
    url = 'https://ftp.hankcs.com/hanlp/tok/coarse_electra_small_20220616_012050.zip'
    下載(url, 儲存目錄=local / 'tok')
    url = 'https://ftp.hankcs.com/hanlp/utils/char_table_20210602_202632.json.zip'
    下載(url, 儲存目錄=local / 'utils')
    url = 'https://ftp.hankcs.com/hanlp/transformers/electra_zh_small_20210706_125427.zip'
    下載(url, 儲存目錄=local / 'transformers')
    url = 'https://ftp.hankcs.com/hanlp/transformers/electra_zh_small_20210706_125427.zip'



def 取學校簡稱(學校名稱):
    import re
    s = 學校名稱
    if s == '宜蘭縣立慈心華德福教育實驗高級中等學校':
        return '慈心華德福實驗高中'
    if s == '慈心高中':
        return '慈心華德福實驗高中'
    if s == '慈心華德福高中':
        return '慈心華德福實驗高中'
    if s == '慈心華德福國小':
        return '慈心華德福實驗高中'
    if s == '慈心華德福':
        return '慈心華德福實驗高中'
    n = s
    pat = r'(.+縣立|.+縣.+[鄉鎮市])(.+)(高中|(國民(中|小|中小)學))(.+分[校班])?'
    if m:=re.match(pat, s):
        n = f'{m[2]}{m[3]}{m[6] if m[6] else ""}'
    n = n.replace('國民中學', '國中').replace('國民小學', '國小')
    n = n.replace('國民中小學', '國中小')
    return n

def 取分校簡稱(學校名稱):
    import re
    s = 學校名稱
    pat = r'.+國小(.+)(分[校班])'
    if m:=re.match(pat, s):
        return m[1] + m[2]
    return s



'花蓮縣政府主管法規資料'
from zhongwen.pandas_tools import 可顯示
from lark import Lark, Transformer
from pathlib import Path
from diskcache import Cache
cache = Cache(Path.home() / 'cache' / 'hllaw')
網址 = 'https://glrs.hl.gov.tw/glrsout/'

def 開啟網頁():
    os.system(f'start {網址}')

async def 抓取(url):
    import logging
    import aiohttp
    import asyncio
    async with aiohttp.ClientSession() as session:
        try:
            async with session.get(url) as response:
                return await response.text()
        except aiohttp.client_exceptions.ClientConnectorError:
            logging.error(f'抓取{url}逾時，停止10秒重試')
            asyncio.sleep(10)
            return await 抓取(url)
 
def 解析法規(內容):
    from bs4 import BeautifulSoup as bs
    import re
    text = 內容
    t = bs(text, 'lxml')
    trs = t.select_one('#content > div.text-con > table').find_all('tr')
    if not trs:
        raise RuntimeError(f'解析法規資訊失敗：{text!r}')
    ths = [tr.th.text.strip().replace('：', '') for tr in trs]
    tds = [tr.td.text.strip() for tr in trs]
    fs = {k:v for k, v in zip(ths, tds)}
    try:
        content = t.select_one('#ctl00_cp_content_divLawContent08').text
        fs['法規內容'] = re.sub(r'\s', '', content.strip())
    except AttributeError:
        fs['法規內容'] = None
    return fs

@可顯示
@cache.memoize(name='爬取花蓮縣法規', expire=30*24*60*60, tag='爬取法規')
def 爬取法規():
    from zhongwen.file import 抓取
    df = 法規連結()
    dfs = []
    for i, r in df.iterrows():
        url = f'https://glrs.hl.gov.tw/glrsout/{r["網址"]}'
        text = 抓取(url)
        l = 解析法規(text)
        dfs.append(l)
    import pandas as pd
    df = pd.DataFrame(dfs)
    return df

@可顯示
def 法規():
    return 法規條文()

def 現行法規():
    '非廢止法規'
    from zhongwen.date import 取日期
    df = 法規().query('not 法規名稱.str.contains("廢")')
    df['修正日期'] = df.修正日期.fillna(df.公發布日)
    df['公發布日'] = df.公發布日.map(取日期)
    df['修正日期'] = df.修正日期.map(取日期)
    return df

async def 取連結(page_no):
    from bs4 import BeautifulSoup as bs
    import pandas as pd
    url = f'https://glrs.hl.gov.tw/glrsout/?page={page_no}'
    text = await 抓取(url)
    t = bs(text, 'lxml')
    trs = t.select('#all > table > tbody > tr')
    links = [tr.select('a')[0] for tr in trs]
    links = [{'名稱':link.text, '網址':link['href']} for link in links]
    df = pd.DataFrame(links)
    return df

async def 爬取法規連結():
    from bs4 import BeautifulSoup as bs
    from io import StringIO
    import pandas as pd
    import asyncio
    import logging
    import re
    url = 'https://glrs.hl.gov.tw/glrsout/'
    text = await 抓取(url)
    logging.info(f'爬取法規連結：{text}')
    pat = r'(\d+)<div class="pageno hidden-xs"'
    if m:=re.search(pat, text):
        page_num = int(m[1])
    else:
        raise RuntimeError('法規主頁無頁數資訊。')
    t = bs(text, 'lxml')
    trs = t.select('#all > table > tbody > tr')
    links = [tr.select('a')[0] for tr in trs]
    links = [{'名稱':link.text, '網址':link['href']} for link in links]
    df = pd.DataFrame(links)
    dfs = await asyncio.gather(*[取連結(page_no) for page_no in range(2, page_num+1)])
    df = pd.concat([df, *dfs])
    return df

@cache.memoize(tag='法規連結', expire=30*24*60*60)
def 法規連結():
    import asyncio
    df = asyncio.run(爬取法規連結())
    return df

def 取法規目錄頁數(html):
    pat = r'(\d+)<div class="pageno hidden-xs"'
    text = html
    if m:=re.search(pat, text):
        return int(m[1])
    else:
        raise RuntimeError('法規主頁無頁數資訊。')

def 取法規頁連結(html):
    text = html
    t = bs(text, 'lxml')
    trs = t.select('#all > table > tbody > tr')
    links = [tr.select('a')[0] for tr in trs]
    links = [{'名稱':link.text, '網址':link['href']} for link in links]
    df = pd.DataFrame(links)
    return df

def 法規分條(法規內容):
    p = Lark.open(Path(__file__).with_suffix('.lark'))
    t = p.parse(法規內容)
    t = 法規分條剖析樹().transform(t).children
    return t

class 法規分條剖析樹(Transformer):
    def LAW_FIRST_PARA(self, tok):
        import re
        from zhongwen.number import 轉數值
        pat = r"第([\d壹貳參肆伍陸柒捌玖拾一二三四五六七八九十]+)條(.*)"
        if m := re.match(pat, tok.value):
            num = 轉數值(m[1])
            p = m[2]
            return (num, p)

    def LAW2_FIRST_PARA(self, tok):
        from zhongwen.number import 轉數值
        import re
        pat = r"([一二三四五六七八九十]+)、(.*)"
        if m := re.match(pat, tok.value):
            num = 轉數值(m[1])
            p = m[2]
            return (num, p)

    def LAW3_FIRST_PARA(self, tok):
        pat = r"([壹貳參肆伍陸柒捌玖拾]+)、(.*)"
        import re
        from zhongwen.number import 轉數值
        if m := re.match(pat, tok.value):
            num = 轉數值(m[1])
            p = m[2]
            return (num, p)

    def PARA(self, tok):
        return tok.value

    def law(self, toks):   
        f, *paras = toks
        return (f[0], ''.join([f[1], *paras]))

    def law2(self, toks):   
        f, *paras = toks
        return (f[0], ''.join([f[1], *paras]))

    def law3(self, toks):   
        f, *paras = toks
        return (f[0], ''.join([f[1], *paras]))

    def law4(self, toks):
        return (1, toks[0].value)

@可顯示
@cache.memoize(name='花蓮縣法規條文', expire=30*24*60*60, tag='法規條文')
def 法規條文(排除廢止者=True):
    from zhongwen.date import 取日期
    import logging
    df = 爬取法規()
    df.rename(columns={'廢止/停止適用日期':'廢止停止適用日期'}, inplace=True)
    df['公發布日'] = df.公發布日.map(取日期)
    df['修正日期'] = df.修正日期.map(取日期)
    df['廢止日期'] = df.廢止日期.map(取日期)
    df['廢止停止適用日期'] = df.廢止停止適用日期.map(取日期)
    if 排除廢止者:
        df = df.query('廢止停止適用日期.isnull()')
    def _法規分條(l):
        try:
            return 法規分條(l)
        except Exception as e:
            logging.error(f'法規分條錯誤：{e}')
            return l
    df['法規分條'] = df.法規內容.map(_法規分條)
    df = df.explode('法規分條')
    def 取條號(t):
        try: return t[0]
        except: return ''
    def 取條文內容(t):
        try: return t[1]
        except: return ''
    df['條號'] = df.法規分條.map(取條號)
    df['條文內容'] = df.法規分條.map(取條文內容)
    df['修正日期'] = df.修正日期.fillna(df.廢止停止適用日期).fillna(df.公發布日)
    df.rename(columns={'修正日期':'異動日期'}, inplace=True)
    return df[['法規名稱', '異動日期', '條號', '條文內容']]

if __name__ == '__main__':
    pass
    import logging
    from zhongwen.pandas_tools import show_html
    # cache.evict('法規條文')
    df = 法規條文()
    df = df.query('法規名稱.str.contains("縣有財產")')
    show_html(df)



from zhongwen.pandas_tools import read_docx, read_fwf, show_html
from zhongwen.pandas_tools import 可顯示

def read_mhtml(mhtml:str):
    from pathlib import Path
    import pandas as pd 
    import email
    
    # 解析 mhtml 文件
    msg = email.message_from_string(mhtml)
    html_content = ""

    # 遍歷 mhtml 文件的內容，提取 HTML 部分
    for part in msg.walk():
        if part.get_content_type() == "text/html":  # 確認是 HTML 部分
            html_content = part.get_payload(decode=True).decode('utf-8')
            break
    # 使用 pandas 讀取 HTML 表格
    return pd.read_html(html_content)

def 取表(字串:str):
    import re
    s = 字串
    pat = r'-[- ]+' # 形式如--- ----- ----
    欄 = None
    if m:=re.search(pat, s):
        欄 = 取欄(m[0])
    ls = s.splitlines() 
    return ls[0] 

def 取字塊(字串:str):
    s = 字串
    p = 0 # 表位置，即 position 簡寫。
    in_block = False
    b = '' # 表字塊，即 block 簡寫。
    bs = []
    start = -1
    end = -1
    for c in s:
        if not c.isspace():
            b += c
            if not in_block:
                start = p
                in_block = True
        else:
            if in_block:
                end = p
                bs.append((b, start, end))
                b = ''
                in_block = False
        p += 字寬(c)
    if in_block:
        end = p
        bs.append((b, start, end))
        b = ''
        in_block = False
    return bs

def 取欄(字串:str):
    import re
    s = 字串
    pat = r'-+'
    cs = []
    for m in re.finditer(pat, s):
        cs.append((m[0], m.start(), m.end()))
    return cs

def 字寬(字元):
    import re
    char = 字元
    return 2 if re.match(r'[\u4e00-\u9fff\u3000-\u303f\uff00-\uffef]', char) else 1



'中文文字處理'
from diskcache import Cache
from pathlib import Path
from functools import lru_cache
import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning) 
warnings.filterwarnings("ignore", category=ResourceWarning) 
cache = Cache(Path.home() / 'cache' / 'text')

@lru_cache
@cache.memoize()
def hantok():
    import hanlp
    return hanlp.load(hanlp.pretrained.tok.COARSE_ELECTRA_SMALL_ZH)

@lru_cache
def 分詞(字串):
    return {w for w in hantok()(字串) if len(w) > 1}

def 去標籤(字串):
    import re
    字串 = 字串.replace('\u200b', '')
    字串 = 字串.replace('\xff', '')
    return re.sub(r'<.*?>', '', 字串)

def 中文詞界(curpos, line):
    import hanlp
    han = hanlp.load(hanlp.pretrained.tok.COARSE_ELECTRA_SMALL_ZH)
    words = (han(line))
    print(words)
    return words

def 臚列(項目):
    "['甲', '乙', '丙'] -> '甲、乙及丙'"
    if type(項目) == list:
        if len(項目) > 1:
            return f"{'、'.join(項目[:-1])}及{項目[-1]}" if len(項目) else ''
        return 項目[0]
    return 項目

def 是否為中文字元(char:str):
    return '\u4e00' <= char <= '\u9fa5'

def 是否為中文符號(char:str):
    import re
    chinese_punctuation_pattern = re.compile(r'[\u3000-\u303F\uFF00-\uFFEF\u2000-\u206F]')
    return bool(re.match(chinese_punctuation_pattern, char))

def 是否為平假名(c:str):
    return "\u3041" <= c <= "\u3096"

def 是否為片假名(c:str):
    return "\u30A1" <= c <= "\u30F6"

def 是否為字元(c:str):
    import string
    # 檢查字元是否為符號
    if c in string.punctuation or 是否為中文符號(c) or c.isspace():
        return False
    return True

全型表 = {i: i + 0xFEE0 for i in range(0x21, 0x7F)}
全型表[0x20] = 0x3000
半型表 = {v: k for k, v in 全型表.items()}

def 轉全型(s):
    return s.translate(全型表)

def 轉半型(s):
    return s.translate(半型表).replace('—', '-')

def 字元切換(string:str):
    '''就大小寫字母，舉如拉丁字母，進行大小寫切換；重碼字母進行校正；中文字母進行簡繁切換；日文字母平片假名切換；符號則為全半型轉換(Todo)。
'''
    from opencc import OpenCC
    if string == '': return string
    def switch_case(c:str):
        if c == '[': return '「'
        if c == ']': return '」'

        if 是否為重碼字(c):
            return 校正中文字(c)

        if 是否為中文字元(c):
            # 簡繁切換
            r = OpenCC('s2t').convert(c)
            if r == c:
                return OpenCC('t2s').convert(c)
            return r
        if 是否為平假名(c):
            return chr(ord(c)-0x3041+0x30A1)
        if 是否為片假名(c):
            return chr(ord(c)-0x30A1+0x3041)
        if c.islower():
            return c.upper()
        else:
            return c.lower()

    if len(string) == 1: return switch_case(string)

    return ''.join(map(switch_case, string))

def 下載倉頡碼對照表():
    from zhongwen.file import 下載
    f = 下載('https://github.com/Jackchows/Cangjie5/raw/master/Cangjie5_TC.txt')
    return f

@cache.memoize()
def 倉頡對照表():
    from pandas import read_csv
    f = 下載倉頡碼對照表()
    df = read_csv(f, encoding='latin1'
                 ,skiprows=12
                 ,sep='\t'
                 ,names=['漢字','倉頡碼','標註']
                 )
    def convert(latin1:str):
        try:
            return latin1.encode('latin1').decode('utf8', errors='replace') 
        except AttributeError:
            return latin1
    df['漢字'] = df.漢字.map(convert)
    d = {}
    for index, row in df.iterrows():
        if not row['漢字'] in d:
            d[row['漢字']] = row['倉頡碼']
    return d

對照表 = 倉頡對照表() # 以模組變數將對照表緩存在記憶體
def 倉頡首碼(char):
    try:
        return 對照表[char][0]
    except KeyError:
        return char

def 首碼搜尋表示式(char, text):
    cs = set()
    for i, c in enumerate(text):
        if c == char or 倉頡首碼(c) == char:
            cs.add(c)
    return f'[{"".join(cs)}]'

def 翻譯(word):
    from pygtrans import Translate
    client = Translate()
    # 翻译句子
    text = client.translate(word, 'zh-tw')
    return text.translatedText

重碼字 ='車路類例宅金易勒精神行煉練復療溜樓例列量料飯旅冷靈禮福利老六數歷里不拉簾來說度便力錄連年爐讀益'
重碼字+='立異理狀'
正體字 ='車路類例宅金易勒精神行煉練復療溜樓例列量料飯旅冷靈禮福利老六數歷里不拉簾來說度便力錄連年爐讀益'
正體字+='立異理狀'

def 是否為重碼字(字串:str):
    return 字串[0] in 重碼字 

def 校正中文字(字串:str):
    t = str.maketrans(重碼字, 正體字)
    return 字串.translate(t)

def 刪空格(n):
    import re
    try: return re.sub(r'\s+', '', n)
    except TypeError: return n

class 萌典尚無定義之字詞(Exception):pass

def 查萌典(字詞):
    from .file import 抓取
    import requests
    import logging
    import json
    url = f'https://www.moedict.tw/{字詞}.json'
    try:
        j = 抓取(url)
        d = json.loads(j)
        hs = d['heteronyms']
        def 取定義(異名):
            h = 異名
            try:
                m = h['bopomofo']
                m += '：'
            except KeyError:
                m = ''
            ds = h['definitions']
            m += ''.join([去標籤(d["def"]) for d in ds])
            return m
        return [取定義(h) for h in hs]
    except (json.JSONDecodeError, requests.exceptions.HTTPError):
        raise 萌典尚無定義之字詞(字詞) 

def 去除字串末句號(字串):
    import re
    return re.sub('。$', '', 字串)

def 公司正名(公司名稱):
    import re
    n = 公司名稱
    n = n.replace("(股)", "股份有限")
    n = re.sub(r'\(下稱.+\)$', '', n)
    return n

def 轉樣式表字串(字串):
    '以將Unicode字串以CSS跳脫字元方式轉成ASCII字串，'
    try:
        original_string = 字串
        unicode_escaped_string = original_string.encode('unicode_escape').decode()
        # print(unicode_escaped_string)
        r'[\u9577\u7a7a]\uff1a\u4e0a\u5e74\u5ea6\u80a1\u5229\u4f30\u8a08\u76f8\u5c0d\u8aa4\u5dee\u905420.40%\uff1b[\u77ed\u591a]\uff1a\u6bdb\u5229\u7387\u905454.27%\uff0c\u71df\u5229\u7387\u905423.48%\uff0c\u71df\u5229\u7387\u8f03\u53bb\u5e74\u540c\u671f\u589e\u52a024.60%\uff0c\u6de8\u5229\u7387\u8f03\u53bb\u5e74\u540c\u671f\u589e\u52a023.19%\uff1b[\u77ed\u7a7a]\uff1a\u672a\u6d41\u901a\u6bd4\u4f8b\u9ad8\u905432.43%\uff0c\u4f30\u8a08\u80a1\u5229\u8870\u6e1b17.47%'
        import re
        pat = r'\\u(\w{4})'
        css_escaped_string = re.sub(pat, r'\\00\1', unicode_escaped_string)
        # print(css_escaped_string)
        return css_escaped_string
    except AttributeError:
        return 字串

def 符號轉半型(s, 輸出錯誤訊息=False):
    import logging
    try:
        return s.translate(str.maketrans('，；：％。、'
                                        ,',;:%.,')
                )
    except AttributeError as e:
        if 輸出錯誤訊息:
            logging.error(f'符號轉半型錯誤{e}')
        return s

def 移除非中文字(text):
    '移除非中文字及符號，可用於表頭標題的正規化。'
    import re
    chinese_pattern = re.compile(r'[^\u4e00-\u9fa5]')  # 匹配非中文字的正則表達式
    return chinese_pattern.sub('', text)

def 是否為空白字元(text):
    raise Exception('請用 str.isspace() 替代！')

def 交談(text):
    import requests 

    headers = {
        "Content-Type": "application/json; charset=utf-8"
    }

    url = 'https://generativelanguage.googleapis.com/v1beta/models/gemini-pro:generateContent?key=AIzaSyDF7loLYLpl7V4XXJzP4uJsOy59vZSOZB4'
    data = '{"contents":[{"parts":[{"text":"'+text+'"}]}]}'
    # data = [{"parts":[{"text":f"{text}"}]}]
    # print(data)
    import json.encoder
    r = requests.post(url, data=data, headers=headers)
    
    return r.json()

環繞開啟符號 = "'\"([{「『（【"
環繞關閉符號 = "'\")]}」』）】"
環繞關閉符號對應 = str.maketrans(環繞開啟符號, 環繞關閉符號)

def 取環繞關閉符號(環繞開啟符號:str):
    return 環繞開啟符號.translate(環繞關閉符號對應)

def 取最近環繞符號(字串:str) -> str:
    最近環繞符號索引 = -1
    最近環繞符號位置 = -1
    for i, c in enumerate(環繞開啟符號):
        p = 字串.rfind(c)
        if p == -1:
            continue
        最近環繞符號位置 = max(最近環繞符號位置, p)
        if 最近環繞符號位置 == p:
            最近環繞符號索引 = i
    if 最近環繞符號位置 == -1:
        raise RuntimeError(f'以下字串尚無環繞符號：{字串}')
    return 環繞開啟符號[最近環繞符號索引] + 環繞關閉符號[最近環繞符號索引]

def 刪除環繞符號(字串:str, 環繞符號='(', 游標位置:int=0) -> str:
    import re
    l = 環繞符號
    r = 取環繞關閉符號(l)
    p = 游標位置
    pat = f"{l}[^{r}]?{r}"
    istart = 0
    iend = 0
    for m in re.finditer(pat, 字串):
        start = m.start()
        end = m.end()
        if start <= p <= end:
            if start > istart:
                istart = start
                iend = end
    if istart != iend:
        return 字串[:start] + 字串[start+1:end-1] + 字串[end:]
    return 字串

def 插入環繞符號(字串:str, 環繞符號='(', 列:int=0, 區間='w') -> str:
    '區間是vim的 motion 命令，如 w、$'
    import re
    match 區間:
        case 'w':
            s = 字串[列-1:]
            pat = r'[a-zA-Z\u4e00-\u9fff]+'
            if m:=re.match(pat, s):
                return f'{字串[:列-1]}{環繞符號}{m[0]}{取環繞關閉符號(環繞符號)}{s[m.end():]}' 

左引號 = "'\"([{「『（【"
右引號 = "'\")]}」』）】"
左引號表 = dict(zip(右引號, 左引號))

from dataclasses import dataclass
@dataclass
class 引用:
    內容:str
    始:int
    迄:int

def 取引用內容(字串:str, 游標位置:int=0):
    s = 字串
    slen = len(s)
    可能右引號 = 右引號
    pos = 游標位置
    rpos = pos
    while len(可能右引號) > 0:
        r = None
        while rpos < slen:
            if s[rpos] in 可能右引號:
                r = s[rpos]
                break
            rpos += 1
        if not r:
            return None
        l = 左引號表[r]
        left = s[:rpos+1]
        lpos = left.rfind(l)
        if pos < lpos:
            可能右引號 = 可能右引號.replace(r, "")
            continue
        return 引用(字串[lpos:rpos+1], lpos, rpos)
    return None

def 查找引號對(expression):
    stack = []
    matches = {}

    for i, char in enumerate(expression):
        if char in "([{":
            stack.append((char, i))
        elif char in ")]}":
            if not stack:
                raise ValueError(f"Unmatched closing {char} at index {i}")
            opening_char, opening_index = stack.pop()
            if (opening_char == '(' and char != ')') or \
               (opening_char == '[' and char != ']') or \
               (opening_char == '{' and char != '}'):
                raise ValueError(f"Unmatched {opening_char} at index {opening_index} with {char} at index {i}")
            matches[opening_index] = i
            matches[i] = opening_index

    if stack:
        remaining_opening = stack.pop()
        raise ValueError(f"Unmatched opening {remaining_opening[0]} at index {remaining_opening[1]}")

    return matches





from pathlib import Path
import unittest
wdir = Path(__file__).parent

class TestPythonDebug(unittest.TestCase):
    def setUp(self):
        test_file = wdir / 'test_file.py'
        test_file.touch()
        self.test_file = test_file

        test_file2 = wdir.parent / 'test_file2.py'
        test_file2.touch()
        self.test_file2 = test_file2

    def test_deploy(self):
        from zhongwen.python_dev import 專案根目錄, 套件名稱, 布署
        import os
        self.assertEqual(專案根目錄(Path(__file__)), Path(__file__).parent.parent)

        self.assertEqual(套件名稱(Path(__file__).parent.parent / 'pyproject.toml'), 'zhongwen')
        
        old_cwd = os.getcwd() 
        os.chdir(str(wdir))
        布署() 
        os.chdir(old_cwd)
        布署(Path(__file__)) 
        self.assertEqual(os.getcwd(), old_cwd)
        with self.assertRaises(FileNotFoundError):
            布署(Path(r'd:\GitHub\vimpython\ftplugin\python.vim'))

    def test_find_testfile(self):
        from zhongwen.python_dev import find_testfile
        import os

        # 中文檔名其測試檔檔名為其檔名前綴【測試】
        test_file = wdir / '測試中文檔.py'
        test_file.touch()
        self.assertEqual(find_testfile(wdir / r'中文檔.py')
                        ,str(wdir / r'測試中文檔.py'))
        test_file.unlink()

        # 檔名前綴為'test_'即為測試檔
        self.assertEqual(find_testfile(wdir / r'test_file.py')
                        ,str(wdir / r'test_file.py'))
        
        # 拉丁文檔名加上前綴'test_'為其測試檔
        # 在同目錄搜尋
        self.assertEqual(find_testfile(wdir / 'file.py')
                        ,str(wdir / f'test_file.py' ))

        # 在父目錄搜尋
        self.assertEqual(find_testfile(wdir / 'file2.py')
                        ,str(wdir.parent / f'test_file2.py' ))

        # 在父目錄搜尋格式為測試前綴目錄名之測試檔
        testpython3 = wdir.parent / 'test_tests.py'
        testpython3.touch()
        self.assertEqual(find_testfile(wdir / 'file3.py')
                        ,str(testpython3))
        testpython3.unlink()

        # test.py 為第二順位測試檔
        testpy = wdir / 'test.py'
        testpy.touch()
        self.assertEqual(find_testfile(wdir / 'file2.py')
                        ,str(wdir.parent / f'test_file2.py'))
        testpy.unlink()
        
    def tearDown(self):
        self.test_file.unlink()
        self.test_file2.unlink()
        
    # def test_至定義(self):
        # code = '''from stock.tifrs_schema import 科目表
        # w = 科目表()
        # '''
        # from jedi import Script
        # import fhopecc.env as env
        # s = Script(code=code, path='test.py')
        # gs = s.goto(2, 6, follow_imports=True)
        # self.assertGreater(len(gs), 0)
        # self.assertEqual(gs[0].module_path, env.workdir / 'stock' / 'tifrs_schema.py')
    
if __name__ == '__main__':
    unittest.main()



from pathlib import Path
from diskcache import Cache
cache = Cache(Path.home() / 'cache' / Path(__file__).stem)

def show_map(gdf):
    '運用 folium 顯示 geopandas 資料框'
    from shapely.geometry import mapping
    from pathlib import Path
    import folium
    import os
    gdf = gdf.to_crs(epsg='4326')
    m = folium.Map(location=[gdf.centroid.y.mean(), gdf.centroid.x.mean()], zoom_start=10)

    for idx, row in gdf.iterrows():
        geojson = mapping(row.geometry)
        folium.GeoJson(geojson).add_to(m)

    html = Path.home() / 'TEMP' / 'output.html'
    m.save(html)
    os.system(f'start {html}')

@cache.memoize(tag='路網')
def 取路網(區域名稱="Yilan County, Taiwan", 類型='drive'):
    """
    :param 類型: 路網類型，例如 'drive'（開車）、'walk'（步行）、'bike'（自行車）等
    :return: OSMnx 圖形物件
    """
    import osmnx as ox
    ox.utils.config(log_console=True, use_cache=True)
    place_name = 區域名稱
    network_type = 類型
    G = ox.graph_from_place(place_name, network_type=network_type)
    return G

def 取節點最短路程(起點, 終點, 路網=None):
    '回傳公尺距離及節點路徑'
    import networkx as nx
    import pandas as pd
    G = 路網
    if not G:
        G = 取路網()
    origin_node = 起點
    destination_node = 終點
    
    try:
        shortest_route = nx.shortest_path(G, origin_node, destination_node, weight='length')
        route_length = nx.shortest_path_length(G, origin_node, destination_node, weight='length')
        if pd.isnull(route_length):
            msg = f"沒有找到從節點 {origin_node} 到節點 {destination_node} 的路徑。"
            raise nx.NetworkXNoPath(msg)
        return route_length, shortest_route
    except nx.NetworkXNoPath:
        msg = f"沒有找到從節點 {origin_node} 到節點 {destination_node} 的路徑。"
        raise nx.NetworkXNoPath(msg)

class 轉換多邊形錯誤(Exception): pass
def 取開放街圖網路圖徵(識別碼):
    from shapely.geometry import Polygon
    import overpy

    api = overpy.Overpass()
    specific_way_id = 識別碼
    query = f"""
    [out:json][timeout:25];
    way({specific_way_id});
    out body;
    >;
    out skel qt;
    """
    result = api.query(query)
    if result.ways:
        way = result.ways[0]
        
        # 創建節點坐標字典
        nodes = {node.id: (float(node.lon), float(node.lat)) for node in result.nodes}
        
        # 獲取多邊形的座標
        coords = [nodes[node.id] for node in way.nodes]
        
        # 確保該 way 是閉合的
        if coords[0] == coords[-1]:
            polygon = Polygon(coords)  # 創建 Shapely Polygon
            return polygon
        else:
            raise 轉換多邊形錯誤("非閉合網路無法轉換成多邊形。")

def 取路口(查詢地點=None):
    '取開放街圖路口'
    import osmnx as ox
    import geopandas as gpd

    # 指定花蓮市的位置名稱
    place_name = "Hualien City, Taiwan"

    # 從 OpenStreetMap 下載花蓮市的道路網絡數據
    G = ox.graph_from_place(place_name, network_type='drive')

    # 提取路口 (交叉點) 的資料
    nodes, edges = ox.graph_to_gdfs(G, nodes=True, edges=True)

    # 篩選出具有多個連接道路的路口 (即交叉點)
    intersections = nodes[nodes['street_count'] > 1]

    def get_intersection_name(node, G):
        "根據路口連接的道路名稱推測路口名稱"
        street_names = set()
        for u, v, key, data in G.edges(node, keys=True, data=True):
            if 'name' in data:
                street_names.add(str(data['name']))
        return "".join(sorted(street_names))

    # 對每個路口推測名稱
    intersections['name'] = intersections.index.map(lambda node: get_intersection_name(node, G))
    return intersections

def 繪路程(起點, 終點, 地圖, 路網=None, 色彩='blue'):
    import folium
    import osmnx as ox
    s, e, m, G = 起點, 終點, 地圖, 路網
    folium.Marker((s.y, s.x)
                 ,icon=folium.DivIcon(
                     html=f'<div style="font-size: 10pt; color: white; background:black">{s.名稱}</div>')
                 ).add_to(m)
    folium.Marker((e.y, e.x)
                 ,icon=folium.DivIcon(
                     html=f'<div style="font-size: 10pt; color: white; background:black">{e.名稱}</div>')
                 ).add_to(m)
    if not G:
        G = 取路網()
    snode = ox.distance.nearest_nodes(G, s.geometry.x, s.geometry.y)
    enode = ox.distance.nearest_nodes(G, e.geometry.x, e.geometry.y)
    d, route = 取節點最短路程(snode, enode, G)
    km = d / 1000  # 將距離轉換為公里

    # 路徑中心節點
    route_center_index = len(route) // 2
    route_center = route[route_center_index]

    # 繪製路徑節點
    route_points = [(G.nodes[node]['y'], G.nodes[node]['x']) for node in route]
    folium.PolyLine(route_points, color=色彩, weight=5, opacity=0.7).add_to(m)

    folium.Marker(
        location=(G.nodes[route_center]['y'], G.nodes[route_center]['x']),
        icon=folium.DivIcon(html=f'<div style="text-align:center;width:48pt; font-size: 8pt; color: white; background:black">{km:.2f}公里</div>')
    ).add_to(m)

@cache.memoize(tag='下載宜蘭縣寺廟圖徵')
def 下載宜蘭縣寺廟圖徵():
    import osmnx as ox
    place_name = "Yilan County, Taiwan"
    tags = {'amenity': 'place_of_worship', "religion": "buddhist"}
    temples = ox.geometries_from_place(place_name, tags)
    return temples



import os
import email
import base64
     
def 另存附件(電子郵件檔, 儲存目錄=None):
    '儲存目錄預設為執行檔目錄'
    if not 儲存目錄:
        from pathlib import Path
        儲存目錄 = Path(__file__).parent
    output_dir = 儲存目錄
    eml_file = 電子郵件檔 
    with open(eml_file, 'rb') as file:
        eml_data = file.read()
    msg = email.message_from_bytes(eml_data)

    for part in msg.walk():
        if part.get_content_disposition() and 'attachment' in part.get_content_disposition():
            filename = part.get_filename()
            if filename:
                attachment_data = part.get_payload(decode=True)
                save_path = os.path.join(output_dir, filename)
                with open(save_path, 'wb') as attachment_file:
                    attachment_file.write(attachment_data)
                print(f'附件儲存為{filename}！')



import unittest

class Test(unittest.TestCase):
    def setUp(self):
        from pathlib import Path
        self.甲 = (Path(__file__).parent / '表測例甲.txt').read_text(encoding='utf-8')

    def test(self):
        from zhongwen.表 import 取表
        字串 = self.甲
        df = 取表(字串)
        print(df)
        self.assertEqual(df.columns[0], '科目')
        
    def test_parse_block(self):
        from zhongwen.表 import 取字塊
        s = '期間                 (月)                      (季)           (最近四季累計)'
        bs = 取字塊(s)
        self.assertEqual(bs[0], ('期間', 0, 4))
        self.assertEqual(bs[1], ('(月)', 21, 25))
        self.assertEqual(bs[-1], ('(最近四季累計)', 62, 76))
        self.assertEqual(len(bs), 4)

    def test_parse_columns(self):
        from zhongwen.表 import 取欄
        分隔線 = '----------  ----------  ----------  ------------  ----------  --------------' 
        cs = 取欄(分隔線) 
        self.assertEqual(len(cs), 6)
        self.assertEqual(cs[0], ('----------', 0, 10))
        self.assertEqual(cs[1], ('----------', 12, 22))

    def test_char_width(self):
        from zhongwen.表 import 字寬
        self.assertEqual(字寬('-'), 1)
        self.assertEqual(字寬('字'), 2)

if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.INFO)
    # unittest.main()
    suite = unittest.TestSuite()
    suite.addTest(Test('test_parse_block'))  # 指定測試
    unittest.TextTestRunner().run(suite)



import unittest

class Test(unittest.TestCase):
    def test(self):
        from zhongwen.法人 import 取學校簡稱, 取分校簡稱
        self.assertEqual(取學校簡稱('宜蘭縣立宜蘭國民中學'), '宜蘭國中')
        self.assertEqual(取學校簡稱('宜蘭縣立岳明國民中小學'), '岳明國中小')
        self.assertEqual(取學校簡稱('宜蘭縣立內城國民中小學化育分校'), '內城國中小化育分校')
        self.assertEqual(取學校簡稱('宜蘭縣縣立南澳高中'), '南澳高中')
        self.assertEqual(取學校簡稱('人文國民中小學'), '人文國中小')
        self.assertEqual(取分校簡稱('龍潭國小匏崙分校'), '匏崙分校') 

if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.INFO)
    logging.getLogger('googleclient').setLevel(logging.CRITICAL)
    logging.getLogger('matplotlib').setLevel(logging.CRITICAL)
    logging.getLogger('faker').setLevel(logging.CRITICAL)
    unittest.main()
    suite = unittest.TestSuite()
    suite.addTest(Test('test'))  # 指定測試
    unittest.TextTestRunner().run(suite)



import unittest
class Test(unittest.TestCase):
    def test(self):
        from zhongwen.消防法規 import 抓取法規
        url = 'https://law.nfa.gov.tw/mobile/law.aspx?LSID=FL005010'
        law = 抓取法規(url)
        from clipboard import copy
        copy(law)
        print(law)

if __name__ == '__main__':
    unittest.main()



import unittest

class Test(unittest.TestCase):

    def test_query_by_content(self):
        from zhongwen.中央法規 import 法規條文, cache
        # cache.clear()
        pass
        # df = 法規條文()
        # df = df.query('法規名稱.str.contains("各類場所消防安全設備設置標準")')
        # df = df.query('條文內容.str.contains("，及")')
        # df = df.query('法規名稱=="消防勤務實施要點"')
        # from fhopecc.pandas_tools import show_html
        # show_html(df)

    def test_law_query(self):
        from zhongwen.法規 import LawQuery
        q = LawQuery('職業安全衛生法第33條')
        self.assertEqual(q.法規名稱, '職業安全衛生法')
        self.assertEqual(q.條號, "33") 
        self.assertEqual(q.關鍵字, None) 

        q = LawQuery('通訊交易解除權合理例外情事適用準則第2條')
        self.assertEqual(q.法規名稱, '通訊交易解除權合理例外情事適用準則')
        self.assertEqual(q.條號, "2") 
        self.assertEqual(q.關鍵字, None) 

        q = LawQuery('職業安全衛生法[合格]')
        self.assertEqual(q.法規名稱, '職業安全衛生法')
        self.assertEqual(q.條號, None) 
        self.assertEqual(q.關鍵字, '合格') 


    def test(self):
        from zhongwen.法規 import 法條查詢, 法條展開, 法條說明

        self.assertEqual(法條查詢('職業安全衛生法第33條').條文內容.iloc[0], 
                '雇主應負責宣導本法及有關安全衛生之規定，使勞工周知。')

        self.assertEqual(法條查詢('職業安全衛生法33').條文內容.iloc[0], 
                '雇主應負責宣導本法及有關安全衛生之規定，使勞工周知。')

        self.assertEqual(法條查詢('職業安全衛生法[合格]').條號.iloc[0], '8')

        self.assertEqual(法條展開('職業安全衛生法33'), 
                '職業安全衛生法第33條規定：「雇主應負責宣導本法及有關安全衛生之規定，使勞工周知。」'
                )
        self.assertEqual(法條說明('職業安全衛生法[合格]')[:10], '職業安全衛生法第8條')

        self.assertEqual(法條展開('危險性機械及設備安全檢查規則3')[:27], 
                '危險性機械及設備安全檢查規則第3條規定：「本規則適用於'
                )
        self.assertEqual(法條展開('直轄市縣市消防機關救護車輛裝備人力配置標準5')[:27], 
                '直轄市縣市消防機關救護車輛裝備人力配置標準第5條規定：'
                )
        self.assertEqual(法條展開('花蓮縣政府組織自治條例1')[:27], 
                '花蓮縣政府組織自治條例第1條規定：「本自治條例依地方行'
                )
        
    def test_law_completion(self):
        from zhongwen.法規 import 法規名稱字首樹, 法規自動完成建議
        self.assertEqual(法規名稱字首樹().keys('漁港法')[0], '漁港法')
        line = '依據證'
        self.assertEqual(法規自動完成建議(line)[0], '證') # 第一個結果是找出的字首
        self.assertEqual(法規自動完成建議(line)[1], '證券交易法') # 名稱越短的法規越重要

if __name__ == '__main__':
    unittest.main()



def show_map(gdf):
    '運用 folium 顯示 geopandas 資料框'
    from shapely.geometry import mapping
    from pathlib import Path
    import folium
    import os
    gdf = gdf.to_crs(epsg='4326')
    m = folium.Map(location=[gdf.centroid.y.mean(), gdf.centroid.x.mean()], zoom_start=10)

    for idx, row in gdf.iterrows():
        geojson = mapping(row.geometry)
        folium.GeoJson(geojson).add_to(m)

    html = Path.home() / 'TEMP' / 'output.html'
    m.save(html)
    os.system(f'start {html}')



import unittest

class Test(unittest.TestCase):
    def test_file_location(self):
        from file import FileLocation
        line = r'File "d:\github\zhongwen\test_file.py", line 9, in test_file_location'

        f = FileLocation(line)
        self.assertEqual(f.路徑, r'd:\github\zhongwen\test_file.py')
        self.assertEqual(f.列, 9)

        line = "g:\我的雲端硬碟\股票分析\公開資訊觀測站爬蟲.py:299: SettingWithCopyWarning:"
        f = FileLocation(line)
        self.assertEqual(f.路徑, r'g:\我的雲端硬碟\股票分析\公開資訊觀測站爬蟲.py')
        self.assertEqual(f.列, 299)

        line = "> g:\我的雲端硬碟\股票分析\股票估值.py(41)"
        f = FileLocation(line)
        self.assertEqual(f.路徑, r'g:\我的雲端硬碟\股票分析\股票估值.py')
        self.assertEqual(f.列, 41)

        line = 'at Object.toBe (jandas.test.js:7:18)'
        f = FileLocation(line)
        self.assertEqual(f.路徑, 'jandas.test.js')
        self.assertEqual(f.列, 7)
        self.assertEqual(f.行, 18)

        line = "const Series = require('./jandas.js').Series"
        f = FileLocation(line)
        self.assertEqual(f.路徑, './jandas.js')
        self.assertEqual(f.列, 0)
        self.assertEqual(f.行, 0)
        
        line = "const Series = require('./jandas.js').Series"
        f = FileLocation(line)
        self.assertEqual(f.路徑, './jandas.js')


    def test_file(self):
        # from zhongwen.file import 抓取, cache
        # cache.clear()
        # url = 'https://glrs.hl.gov.tw'
        # page = 抓取(url)
        # print(page.find_elements_by_css_selector('table'))
        # self.assertEqual(page[:5], "<html")
        pass

if __name__ == '__main__':
    unittest.main()




'Python 開發工具'
from pathlib import Path
import logging
logger = logging.getLogger(Path(__file__).stem)

def 專案根目錄(檔案:Path):
    p = 檔案
    while not(p.parent / 'pyproject.toml').exists():
        if p == p.parent:
            raise FileNotFoundError(f'【{檔案}】無專案目錄！')
        p = p.parent
    return p.parent

def 套件名稱(f):
    from toml import load
    t = load(f)
    return t['project']['name']

def 布署(檔案:Path=None):
    import os
    if 檔案:
       r = 專案根目錄(檔案)
       old_cwd = os.getcwd()
       os.chdir(str(r))
       布署()
       os.chdir(old_cwd)
    else:
        try:
            n = 套件名稱('pyproject.toml') 
            cmd =  f'del dist\* && py -m build && twine upload dist\* && '
            cmd += f'python -m pip install {n} -U && '
            cmd += f'python -m pip install {n} -U'
            os.system(cmd)
        except FileNotFoundError:
            logger.error(f'當前目錄 {os.getcwd()} 無 pyproject.toml 檔案！')

def 至():
    from zhongwen.file import FileLocation
    import vim
    line = vim.eval("getline('.')")
    錯誤位置 = FileLocation(line)
    vim.command(f"e +{錯誤位置.列} {錯誤位置.路徑}")

def find_testfile(f:Path, debug=True):
    '依據路徑推論測試檔位置'
    import re
    if not isinstance(f, Path): 
        f = Path(f)

    def 是否為中文檔名(檔名):
        from zhongwen.text import 是否為中文字元
        for c in 檔名:
            if 是否為中文字元(c): return True
        return False
    
    測試檔前綴 = '測' if 是否為中文檔名(f.name) else 'test_' 

    test = f
    pat = 測試檔前綴
    if m:=re.match(pat, test.name) and test.exists():
        return str(test) 

    test = f.parent.parent / 'tests' / f'{測試檔前綴}{f.name}'
    if test.exists(): return str(test) 

    test = f.parent / f'{測試檔前綴}{f.name}'
    if test.exists(): return str(test) 

    test = f.parent / 'tests' / f'{測試檔前綴}{f.name}'
    if test.exists(): return str(test) 

    test = f.parent.parent / f'{測試檔前綴}{f.name}'
    if test.exists(): return str(test) 
     
    test = f.parent / 'test.py'
    if test.exists(): return str(test) 

    test = f.parent.parent / f'{測試檔前綴}{f.parent.name}.py'
    if test.exists(): return str(test) 
    
    raise FileNotFoundError(f'{f.name}尚無測試檔！')

def 光標物件():
    '光標處Python物件'
    import vim, jedi
    f = vim.eval("expand('%')")
    c = '\n'.join(vim.current.buffer)
    script = jedi.Script(code=c, path=f)
    _a, l, c, _a, _a = map(lambda s: int(s), vim.eval('getcursorcharpos()'))
    try:
        return script.goto(l, c, follow_imports=True)[0]
    except IndexError: None

def 至定義():
    '至光標處Python物件之定義'
    import vim
    r = 光標物件()
    if r:
        p = r.module_path
        l = r.line
        c = r.column
        vim.command(f"e +{l} {p}")

def escape_vim_special_chars(text):
    import vim
    escaped_text = vim.eval('escape("' + text + '", "#")')
    return escaped_text

def 說明():
    import vim
    r = 光標物件()
    if r:
        m = r.docstring()
        m = m.split("\n")
        import re
        m = [_m.replace('"', '＂').replace("'", '＇') for _m in m]
        m = [f"'{_m}'" for _m in m]
        m = ','.join(m)
        m = f"[{m}]"
        # print(f"{m!r}")
        # m = escape_vim_special_chars(m)
        vim.command(f"call popup_atcursor({m}, {{}})")

def 當前目錄加入模組查找路徑():
    '因 Python 係命令提示字元之子行程，無法更改其環境變數，須重啟。'
    import os
    pythonpath = ";".join([str(Path(os.getcwd())), os.environ['PYTHONPATH']])
    os.system(f'setx PYTHONPATH {pythonpath}')

def setgit():
    import os
    cmd = 'git config --global core.quotepath off'
    os.system(cmd)

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--deploy2pypi"
                       ,help="布署至PyPI，執行目錄下應有 pyproject.toml 檔。"
                       ,action='store_true')
    parser.add_argument("-a", "--add2pythonpath"
                       ,help="當前目錄加入模組查找路徑。"
                       ,action='store_true')
    parser.add_argument("--setgit"
                       ,help="設定 git 環境。"
                       ,action='store_true')
    args = parser.parse_args()
    if args.deploy2pypi:
        布署()
    elif args.add2pythonpath:
        當前目錄加入模組查找路徑()
    elif args.setgit:
        setgit()
    else:
        pass



from unittest.mock import patch
from zhongwen.date import 今日
import unittest

class Test(unittest.TestCase):
    def setUp(self):
        from zhongwen.batch_data import 取資料庫
        from zhongwen.date import 取日期
        from pathlib import Path
        import pandas as pd
        data = [['1101', '合併', pd.NA, '110', 'text']
               ,['1102', '合併', '100', '110', 'text']
               ]
        self.損益表集甲 = pd.DataFrame(data
                          ,columns=['股票代號', '財報種類', '毛利', '營利', '其他'])

        data = [['1101', '合併', '100', '110', 'text']
               ,['1102', '合併', '101', '111', 'text']
               ]
        self.損益表集乙 = pd.DataFrame(data
                          ,columns=['股票代號', '財報種類', '毛利', '營利', '其他'])

        data = [['1101', '合併', '100', '110', 'text']
               ,['1102', '合併', '101', '111', 'text']
               ,['1104', '個別', '101', '111', 'text']
               ]
        self.損益表集丙 = pd.DataFrame(data
                          ,columns=['股票代號', '財報種類', '毛利', '營利', '其他'])


        self.測試資料庫路徑 = Path.home() / 'TEMP' / '測試資料庫'

        self.原始測試資料庫路徑 = Path.home() / 'TEMP' / '原始測試資料庫'

    def tearDown(self):
        from zhongwen.batch_data import 取資料庫
        取資料庫(self.測試資料庫路徑).close()
        self.測試資料庫路徑.unlink()
        取資料庫(self.原始測試資料庫路徑).close()
        self.原始測試資料庫路徑.unlink()

    def test_date_converter(self):
        from zhongwen.date import 取日期
        from zhongwen.batch_data import 轉日期字串
        self.assertEqual(轉日期字串(取日期('1979.7.29')), '1979-7-29')

    def test_sqlite_accessor(self):
        '測試讀寫SQLite功能'
        from zhongwen.batch_data import 取資料庫檔路徑, 批次寫入, 批次讀取
        from zhongwen.batch_data import 取資料表內容, 取資料庫, 取原始資料表內容
        from zhongwen.batch_data import 自原始資料庫重建
        from zhongwen.date import 取日期
        from pathlib import Path 

        self.assertEqual(取資料庫檔路徑(取資料庫(self.測試資料庫路徑)), self.測試資料庫路徑)

        # 測試指定欄位
        批次寫入(self.損益表集甲
                ,取日期('113.9.30')
                ,'財報日期'
                ,'損益表'
                ,取資料庫(self.測試資料庫路徑)
                ,['股票代號', '財報種類', '毛利', '營利'])

        df = 取資料表內容(self.測試資料庫路徑, '損益表')
        self.assertEqual(df.shape, (2, 5))
        self.assertEqual(df.iloc[-1].財報日期, ('2024-9-30'))

        df = 取資料表內容(self.原始測試資料庫路徑, '損益表')
        self.assertEqual(df.shape, (2, 6))
        self.assertEqual(df.iloc[-1].財報日期, ('2024-9-30'))
        self.assertEqual(df.iloc[-1].其他, 'text')

        # 測試增加批次資料
        批次寫入(self.損益表集乙
                ,取日期('113.12.31')
                ,'財報日期'
                ,'損益表'
                ,取資料庫(self.測試資料庫路徑)
                ,['股票代號', '財報種類', '毛利', '營利'])

        df = 取資料表內容(self.測試資料庫路徑, '損益表')
        self.assertEqual(df.shape, (4, 5))
        df = 取資料表內容(self.原始測試資料庫路徑, '損益表')
        self.assertEqual(df.shape, (4, 6))
        
        # 測試覆寫批次功能，須刪除舊批次資料後，寫入新批次資料
        批次寫入(self.損益表集丙 
                ,取日期('113.9.30')
                ,'財報日期'
                ,'損益表'
                ,取資料庫(self.測試資料庫路徑)
                ,['股票代號', '財報種類', '毛利', '營利'])
        批次寫入(self.損益表集丙 
                ,取日期('113.9.30')
                ,'財報日期'
                ,'損益表'
                ,取資料庫(self.測試資料庫路徑)
                ,['股票代號', '財報種類', '毛利', '營利'])
        df = 取資料表內容(self.測試資料庫路徑, '損益表')
        df = df.query('財報日期=="2024-9-30"')
        self.assertEqual(df.shape, (3, 5))

        df = 取原始資料表內容(self.測試資料庫路徑, '損益表')
        self.assertEqual(df.shape, (5, 6))
        df = df.query('財報日期=="2024-9-30"')
        self.assertEqual(df.shape, (3, 6))

        df = 批次讀取(取日期('113.9.30')
                     ,'財報日期'
                     ,'損益表'
                     ,取資料庫(self.測試資料庫路徑)
                     )
        self.assertEqual(df.shape, (3, 6))
        self.assertEqual(df.loc[0, '股票代號'], '1101')
        self.assertEqual(df.loc[0, '毛利'], '100')

        批次寫入(self.損益表集甲
                ,取日期('113.6.30')
                ,'財報日期'
                ,'損益表'
                ,取資料庫(self.測試資料庫路徑)
                ,指定欄位=['股票代號', '財報種類', '毛利', '營利'
                          ,'稅前淨利', '淨利'])

        df = 取原始資料表內容(self.測試資料庫路徑, '損益表')
        self.assertEqual(df.shape, (7, 6))

        df = 批次讀取(取日期('113.6.30')
                     ,'財報日期'
                     ,'損益表'
                     ,取資料庫(self.測試資料庫路徑)
                     )
        self.assertEqual(df.loc[0, '股票代號'], '1101')
        self.assertEqual(df.loc[0, '毛利'], None)

        # 測試自原始資料料表重建
        自原始資料庫重建(self.測試資料庫路徑, '損益表', 
                        ['財報日期', '股票代號', '財報種類'
                        ,'其他'])
        df = 取資料表內容(self.測試資料庫路徑, '損益表')
        self.assertEqual(df.shape, (7, 4))
 
    @patch('zhongwen.date.今日')
    def test(self, 仿今日):
        from zhongwen.batch_data import 解析更新期限
        from zhongwen.date import 取日期, 民國年月
        仿今日.return_value = 取日期('113.3.14')

        時期, 期限, 資訊 = 解析更新期限()
        self.assertEqual(時期, 取日期('113.2.29'))
        self.assertEqual(期限, 取日期('113.3.10'))
        self.assertEqual(資訊, f'113年3月14日應公布113年2月資訊。')

        時期, 期限, 資訊 = 解析更新期限('次月底前')
        self.assertEqual(時期, 取日期('113.2.29'))
        self.assertEqual(期限, 取日期('113.3.31'))
        self.assertEqual(資訊, f'113年3月14日應公布113年2月資訊。')

        時期, 期限, 資訊 = 解析更新期限('次年3個月內')
        self.assertEqual(時期, 取日期('112.12.31'))
        self.assertEqual(期限, 取日期('113.3.31'))
        self.assertEqual(資訊, f'113年3月14日應公布112年資訊。')

        仿今日.return_value = 取日期('113.4.6')
        時期, 期限, 資訊 = 解析更新期限('次季45日內')
        self.assertEqual(時期, 取日期('113.3.31'))
        self.assertEqual(期限, 取日期('113.5.15'))
        self.assertEqual(資訊, f'113年4月6日應公布113年第1季資訊。')

        仿今日.return_value = 取日期('113.7.23')
        時期, 期限, 資訊 = 解析更新期限('次季45日內')
        self.assertEqual(時期, 取日期('113.6.30'))
        self.assertEqual(期限, 取日期('113.8.14'))
        self.assertEqual(資訊, f'113年7月23日應公布113年第2季資訊。')

        仿今日.return_value = 取日期('113.7.23')
        時期, 期限, 資訊 = 解析更新期限('次季2個月內')
        self.assertEqual(時期, 取日期('113.6.30'))
        self.assertEqual(期限, 取日期('113.8.31'))
        self.assertEqual(資訊, f'113年7月23日應公布113年第2季資訊。')

        仿今日.return_value = 取日期('113.10.10')
        時期, 期限, 資訊 = 解析更新期限('次季45日內')
        self.assertEqual(時期, 取日期('113.9.30'))
        self.assertEqual(期限, 取日期('113.11.14'))
        self.assertEqual(資訊, f'113年10月10日應公布113年第3季資訊。')

        仿今日.return_value = 取日期('113.4.22')
        時期, 期限, 資訊 = 解析更新期限('財報')
        self.assertEqual(時期, 取日期('113.3.31'))
        self.assertEqual(期限, 取日期('113.5.30'))
        self.assertEqual(資訊, f'113年4月22日應公布113年第1季資訊。')

if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.DEBUG)
    unittest.main()
